﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.IO;
using System.Diagnostics;

namespace Merthsoft.CALCnet {
	/// <summary>
	/// The main CALCnet class. Handles all the sending and receiving,
	/// </summary>
	public class CALCnet {
		/// <summary>
		/// Gets the name of the meta-hub
		/// </summary>
		public string HubName { get; protected set; }
		
		/// <summary>
		/// Gets the local name of this hub
		/// </summary>
		public string LocalName { get; protected set; }

		/// <summary>
		/// Gets the server to connect on
		/// </summary>
		public string Server { get; protected set; }

		/// <summary>
		/// Gets the port of the server
		/// </summary>
		public int Port { get; protected set; }

		/// <summary>
		/// Gets the calcID.
		/// </summary>
		public byte[] CalcID { get; protected set; }

		private Socket Gcn { get; set; }

		/// <summary>
		/// Constructs a new calcnet object, but does not connect.
		/// </summary>
		/// <param name="hubName">The name of the meta-hub</param>
		/// <param name="localName">The name of the local hub</param>
		/// <param name="server">The server on which to connect</param>
		/// <param name="port">The port of the server</param>
		/// <param name="calcID">The calc ID</param>
		public CALCnet(string hubName, string localName, string server, int port, ICollection<byte> calcID) {
			HubName = hubName;
			LocalName = localName;
			Server = server;
			Port = port;
			CalcID = (byte[])calcID;

			Gcn = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
		}

		/// <summary>
		/// Connects to the server.
		/// </summary>
		public void Connect() {
			if (Gcn.Connected) {
				Gcn.Disconnect(true);
			}
			Gcn.Connect(Server, Port);
			SendPacket(new JoinPacket(HubName, LocalName));
			SendPacket(new CalcPacket(CalcID));
		}

		/// <summary>
		/// Sends the packet to the meta-hub
		/// </summary>
		/// <param name="packet">The packet to send</param>
		/// <returns>True if sending succeeded, false otherwise</returns>
		public bool SendPacket(CALCnetPacket packet) {
			byte[] bytes = packet.ToFormedPacket();

			Debug.Write(string.Format("< [{2}] {0} {1} ", packet.Type, packet.Size, DateTime.Now.ToString("hh:mm:ss")));
			foreach (byte b in packet.Payload) {
				Debug.Write(string.Format("{0:X2}", b));
			}
			Debug.WriteLine("");

			int sent = 0;
			int total = 0;
			while (total < bytes.Length) {
				try { sent = Gcn.Send(bytes); } catch { return false; }
				if (sent == 0) {
					return false;
				} else {
					total += sent;
				}
			}

			return true;
		}

		/// <summary>
		/// Sends a broadcast packet with the given data, using the current calcID
		/// </summary>
		/// <param name="data">The data to send along with the broadcast</param>
		/// <returns>True if sending succeeded, false otherwise</returns>
		public bool SendBroadcast(ICollection<byte> data) {
			BroadcastPacket packet = new BroadcastPacket(CalcID, data);
			return SendPacket(packet);
		}

		/// <summary>
		/// Sends a direct packet with the given data, using the current calcID, to the given calcID
		/// </summary>
		/// <param name="toCalcID">The calc to send the data to</param>
		/// <param name="data">The data to send along with the broadcast</param>
		/// <returns>True if sending succeeded, false otherwise</returns>
		public bool SendDirect(ICollection<byte> toCalcID, ICollection<byte> data) {
			DirectPacket packet = new DirectPacket(CalcID, toCalcID, data);
			return SendPacket(packet);
		}

		/// <summary>
		/// Blocks until a packet is reviewed from the meta-hub
		/// </summary>
		/// <returns>The packet sent from the meta-hub</returns>
		public CALCnetPacket GetData() {
			CALCnetPacket packet = new CALCnetPacket();

			byte[] header = new byte[3];
			try { Gcn.Receive(header, 3, SocketFlags.Partial); } catch { return null; }
			//packet.Size = header[0] + (header[1] << 8);
			packet.Size = header.ToUShort(0, !BitConverter.IsLittleEndian);
			packet.Type = (CALCnetPacket.PacketType)(char)header[2];
			byte[] buffer = new byte[packet.Size];
			int rcvd = 0;
			int total = 0;
			while (total < packet.Size) {
				try { rcvd = Gcn.Receive(buffer, packet.Size, SocketFlags.Partial); } catch { throw; }
				if (rcvd == 0) {
					return null;
				} else {
					packet.Payload.AddRange(buffer);
					total += rcvd;
				}
			}

			Debug.Write(string.Format("> [{2}] {0} {1} ", packet.Type, packet.Size, DateTime.Now.ToString("hh:mm:ss")));
			foreach (byte b in packet.Payload) {
				Debug.Write(string.Format("{0:X2}", b));
			}
			Debug.WriteLine("");

			return packet;
		}

		/// <summary>
		/// Returns a random calcID, guaranteed to not conflict with any real calculator IDs
		/// </summary>
		/// <returns>A 5-byte array</returns>
		public static byte[] RandomCalcID() {
			Random r = new Random();
			byte[] calcID = new byte[5];
			do {
				r.NextBytes(calcID);
			} while (calcID[0] == 0x04 || calcID[0] == 0x0A);
			return calcID;
		}
	}
}
